# -*- encoding: utf-8 -*-
"""Representation of users having an evergreen role."""
from typing import TYPE_CHECKING, Any, Dict

from evergreen.base import _BaseEvergreenObject, evg_attrib

if TYPE_CHECKING:
    from evergreen.api import EvergreenApi


class UsersForRole(_BaseEvergreenObject):
    """Representation of a list of users having an evergreen role."""

    users = evg_attrib("users")

    def __init__(self, json: Dict[str, Any], api: "EvergreenApi") -> None:
        """Create an instance of a UsersForRole object."""
        super(UsersForRole, self).__init__(json, api)



# -*- encoding: utf-8 -*-
"""Evergreen representation of a user's permissions."""
from __future__ import absolute_import

from enum import Enum
from typing import TYPE_CHECKING, Any, Dict

from evergreen.base import _BaseEvergreenObject, evg_attrib

if TYPE_CHECKING:
    from evergreen.api import EvergreenApi


class PermissionableResourceType(str, Enum):
    """Represents resource types that a user can be granted permissions to."""

    PROJECT = "project"
    DISTRO = "distro"
    SUPERUSER = "superuser"


class RemovablePermission(str, Enum):
    """Represents a permission that can be removed from a user."""

    PROJECT = "project"
    DISTRO = "distro"
    SUPERUSER = "superuser"
    ALL = "all"


class ResourceTypePermissions(_BaseEvergreenObject):
    """Representation of a user's permissions on resources of a specific type."""

    resource_type = evg_attrib("type")
    permissions = evg_attrib("permissions")

    def __init__(self, json: Dict[str, Any], api: "EvergreenApi") -> None:
        """
        Create an instance of a resource type permission set.

        :param json: JSON of resource type permissions.
        :param api: Evergreen API.
        """
        super(ResourceTypePermissions, self).__init__(json, api)



"""Representation of project aliases."""
from typing import TYPE_CHECKING, Any, Dict, List

from evergreen.base import _BaseEvergreenObject, evg_attrib

if TYPE_CHECKING:
    from evergreen.api import EvergreenApi


class DisplayTaskAlias(_BaseEvergreenObject):
    """Representation of a DisplayTask in an alias."""

    name = evg_attrib("Name")
    execution_tasks = evg_attrib("ExecutionTasks")

    def __init__(self, json: Dict[str, Any], api: "EvergreenApi") -> None:
        """
        Create an instance of a display task alias.

        :param json: json representing a display task alias.
        :param api: instance of evergreen api object.
        """
        super().__init__(json, api)


class VariantAlias(_BaseEvergreenObject):
    """Representation of an alias for a particular build variant."""

    variant = evg_attrib("Variant")
    tasks = evg_attrib("Tasks")

    def __init__(self, json: Dict[str, Any], api: "EvergreenApi") -> None:
        """
        Create an instance of a variant alias.

        :param json: json representing variant alias.
        :param api: instance of evergreen api object.
        """
        super().__init__(json, api)

    @property
    def display_tasks(self) -> List[DisplayTaskAlias]:
        """Get a list of display tasks for the alias."""
        if not self.json["DisplayTasks"]:
            return []
        return [DisplayTaskAlias(dt, self._api) for dt in self.json["DisplayTasks"]]



# -*- encoding: utf-8 -*-
"""Metrics for an evergreen version."""
from __future__ import absolute_import, division

from datetime import datetime
from typing import TYPE_CHECKING, Callable, Dict, List, Optional

from structlog import get_logger

if TYPE_CHECKING:
    from evergreen.build import Build
    from evergreen.metrics.buildmetrics import BuildMetrics
    from evergreen.version import Version

LOGGER = get_logger(__name__)


class VersionMetrics(object):
    """Metrics about an evergreen version."""

    def __init__(self, version: "Version") -> None:
        """
        Create an instance of version metrics.

        :param version: Version to analyze.
        """
        self.version = version

        self.total_processing_time = 0
        self.task_success_count = 0
        self.task_failure_count = 0
        self.task_timeout_count = 0
        self.task_system_failure_count = 0

        self._create_times: List[datetime] = []
        self._start_times: List[datetime] = []
        self._finish_times: List[datetime] = []

        self.build_metrics: List[BuildMetrics] = []
        self.build_list: Optional[List[Build]] = None

    def calculate(self, task_filter_fn: Optional[Callable] = None) -> "VersionMetrics":
        """
        Calculate metrics for the given build.

        :param task_filter_fn: function to filter tasks included for metrics, should accept a task
                               argument.
        :returns: self.
        """
        self.build_list = self.version.get_builds()
        for build in self.build_list:
            self._count_build(build, task_filter_fn)

        return self

    @property
    def create_time(self) -> Optional[datetime]:
        """
        Time the first task of the version was created.

        :return: Time first task was created.
        """
        if self._create_times:
            return min(self._create_times)
        return None

    @property
    def start_time(self) -> Optional[datetime]:
        """
        Time first task of version was started.

        :return: Time first task was started.
        """
        if self._start_times:
            return min(self._start_times)
        return None

    @property
    def end_time(self) -> Optional[datetime]:
        """
        Time last task of version was completed.

        :return: Time last task was completed.
        """
        if self._finish_times:
            return max(self._finish_times)
        return None

    @property
    def makespan(self) -> Optional[float]:
        """
        Wall clock duration of version.

        :return: Duration of version in seconds.
        """
        if self.start_time and self.end_time:
            return (self.end_time - self.start_time).total_seconds()
        return None

    @property
    def wait_time(self) -> Optional[float]:
        """
        Wall clock duration until version was started.

        :return: Duration until version was started in seconds.
        """
        if self.start_time and self.create_time:
            return (self.start_time - self.create_time).total_seconds()
        return None

    @property
    def total_tasks(self) -> int:
        """
        Get the total tasks in the version.

        :return: total tasks
        """
        return self.task_success_count + self.task_failure_count

    @property
    def pct_tasks_success(self) -> float:
        """
        Get the percentage of successful tasks.

        :return: Percentage of successful tasks.
        """
        return self._percent_tasks(self.task_success_count)

    @property
    def pct_tasks_failure(self) -> float:
        """
        Get the percentage of failure tasks.

        :return: Percentage of failure tasks.
        """
        return self._percent_tasks(self.task_failure_count)

    @property
    def pct_tasks_timeout(self) -> float:
        """
        Get the percentage of timeout tasks.

        :return: Percentage of timeout tasks.
        """
        return self._percent_tasks(self.task_timeout_count)

    @property
    def pct_tasks_system_failure(self) -> float:
        """
        Get the percentage of system failure tasks.

        :return: Percentage of system failure tasks.
        """
        return self._percent_tasks(self.task_system_failure_count)

    def _percent_tasks(self, n_tasks: int) -> float:
        """
        Calculate the percent of n_tasks out of total.

        :param n_tasks: Number of tasks to calculate percent of.
        :return: percentage n_tasks is out of total tasks.
        """
        if self.total_tasks == 0:
            return 0

        return n_tasks / self.total_tasks

    def _count_build(self, build: "Build", task_filter_fn: Optional[Callable]) -> None:
        """
        Add stats for the given build to the metrics.

        :param task_filter_fn: function to filter tasks included for metrics, should accept a task
                               argument.
        :param build: Build to add.
        """
        log = LOGGER.bind(build_id=build.id)
        if build.activated:
            log.debug("Processing metrics for build")
            # If all tasks have been undispatched there is no data.
            if not build.tasks or build.status_counts.undispatched == len(build.tasks):
                log.warning("Build had no tasks or all tasks undispatched")
                return

            build_metrics = build.get_metrics(task_filter_fn)
            if build_metrics:
                self.build_metrics.append(build_metrics)

                self.total_processing_time += build_metrics.total_processing_time
                self.task_success_count += build_metrics.success_count
                self.task_failure_count += build_metrics.failure_count
                self.task_timeout_count += build_metrics.timed_out_count
                self.task_system_failure_count += build_metrics.system_failure_count

                if build_metrics.create_time:
                    self._create_times.append(build_metrics.create_time)

                if build_metrics.start_time:
                    self._start_times.append(build_metrics.start_time)

                if build_metrics.end_time:
                    self._finish_times.append(build_metrics.end_time)

    def as_dict(self, include_children: bool = False) -> Dict:
        """
        Provide a dictionary representation.

        :param include_children: Include child build tasks in dictionary.
        :return: Dictionary of metrics.
        """
        metric = {
            "version": self.version.version_id,
            "total_processing_time": self.total_processing_time,
            "task_total": self.total_tasks,
            "task_success_count": self.task_success_count,
            "task_pct_success": self.pct_tasks_success,
            "task_failure_count": self.task_failure_count,
            "task_pct_failed": self.pct_tasks_failure,
            "task_timeout_count": self.task_timeout_count,
            "task_system_failure_count": self.task_system_failure_count,
        }

        if include_children:
            metric["build_metrics"] = [bm.as_dict(include_children) for bm in self.build_metrics]

        return metric

    def __str__(self) -> str:
        """
        Create string version of metrics.

        :return: String version of metrics.
        """
        process_time_min = self.total_processing_time / 60 if self.total_processing_time else 0
        makespan = self.makespan if self.makespan else 0
        makespan_min = self.makespan / 60 if self.makespan else 0
        waittime = self.wait_time if self.wait_time else 0
        waittime_min = self.wait_time / 60 if self.wait_time else 0

        return """Version Id: {version}
        Total Processing Time: {total_processing_time:.2f}s ({total_processing_time_min:.2f}m)
        Makespan: {makespan:.2f}s ({makespan_min:.2f}m)
        Wait Time: {waittime:.2f}s ({waittime_min:.2f}m)
        Total Tasks: {task_total}
        Successful Tasks: {task_success_count} ({task_pct_success:.2%})
        Failed Tasks: {task_failure_count} ({task_pct_failed:.2%})
        Timed Out Tasks: {task_timeout_count} ({task_pct_timed_out:.2%})
        System Failure Tasks: {task_system_failure_count} ({task_pct_system_failure:.2%})
        """.format(
            version=self.version.version_id,
            total_processing_time=self.total_processing_time,
            total_processing_time_min=process_time_min,
            makespan=makespan,
            makespan_min=makespan_min,
            waittime=waittime,
            waittime_min=waittime_min,
            task_total=self.total_tasks,
            task_success_count=self.task_success_count,
            task_pct_success=self.pct_tasks_success,
            task_failure_count=self.task_failure_count,
            task_pct_failed=self.pct_tasks_failure,
            task_timeout_count=self.task_timeout_count,
            task_pct_timed_out=self.pct_tasks_timeout,
            task_system_failure_count=self.task_system_failure_count,
            task_pct_system_failure=self.pct_tasks_system_failure,
        ).rstrip()



# -*- encoding: utf-8 -*-
"""Metrics for Evergreen builds."""
from __future__ import absolute_import, division

from collections import defaultdict
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Callable, Dict, List, Optional

from structlog import get_logger

from evergreen.errors.exceptions import ActiveTaskMetricsException
from evergreen.task import StatusScore

if TYPE_CHECKING:
    from evergreen.build import Build
    from evergreen.task import Task


LOGGER = get_logger(__name__)


class BuildMetrics(object):
    """Metrics about an evergreen build."""

    def __init__(self, build: "Build") -> None:
        """
        Create an instance of build metrics.

        :param build: Build to analyze.
        """
        self.build = build

        self.success_count = 0
        self.failure_count = 0
        self.undispatched_count = 0
        self.timed_out_count = 0
        self.system_failure_count = 0

        self.display_success_count = 0
        self.display_failure_count = 0
        self.display_undispatched_count = 0
        self.display_timed_out_count = 0
        self.display_system_failure_count = 0

        self.total_processing_time = 0

        self._create_times: List[datetime] = []
        self._start_times: List[datetime] = []
        self._finish_times: List[datetime] = []

        self.task_list: List[Task] = []

        self._display_map: Dict[str, List[Task]] = defaultdict(list)

    def calculate(self, task_filter_fn: Optional[Callable] = None) -> "BuildMetrics":
        """
        Calculate metrics for the given build.

        :param task_filter_fn: function to filter tasks included for metrics, should accept a task
                               argument.
        :returns: self.
        """
        all_tasks = self.build.get_tasks()
        filtered_task_list = all_tasks
        if task_filter_fn:
            filtered_task_list = [task for task in filtered_task_list if task_filter_fn(task)]

        self.task_list = filtered_task_list

        # We want to track display tasks, but not use them for metrics since they are just
        # containers to other tasks.
        filtered_task_list = [task for task in self.task_list if not task.display_only]
        for task in filtered_task_list:
            self._count_task(task)
        self._count_display_tasks()
        return self

    @property
    def total_tasks(self) -> int:
        """
        Get the total tasks in the build.

        :return: total tasks.
        """
        return self.success_count + self.failure_count + self.undispatched_count

    @property
    def pct_tasks_success(self) -> float:
        """
        Get the percentage of successful tasks.

        :return: Percentage of successful tasks.
        """
        return self._percent_tasks(self.success_count)

    @property
    def pct_tasks_undispatched(self) -> float:
        """
        Get the percentage of undispatched tasks.

        :return: Percentage of undispatched tasks.
        """
        return self._percent_tasks(self.undispatched_count)

    @property
    def pct_tasks_failed(self) -> float:
        """
        Get the percentage of failed tasks.

        :return: Percentage of failed tasks.
        """
        return self._percent_tasks(self.failure_count)

    @property
    def pct_tasks_timed_out(self) -> float:
        """
        Get the percentage of timeout tasks.

        :return: Percentage of timeout tasks.
        """
        return self._percent_tasks(self.timed_out_count)

    @property
    def pct_tasks_system_failure(self) -> float:
        """
        Get the percentage of system failure tasks.

        :return: Percentage of system failure tasks.
        """
        return self._percent_tasks(self.system_failure_count)

    @property
    def total_display_tasks(self) -> int:
        """
        Get the total display tasks in the build.

        :return: total display tasks.
        """
        return (
            self.display_success_count
            + self.display_failure_count
            + self.display_undispatched_count
        )

    @property
    def pct_display_tasks_success(self) -> float:
        """
        Get the percentage of successful display tasks.

        :return: Percentage of successful display tasks.
        """
        return self._percent_display_tasks(self.display_success_count)

    @property
    def pct_display_tasks_undispatched(self) -> float:
        """
        Get the percentage of undispatched display_tasks.

        :return: Percentage of undispatched display_tasks.
        """
        return self._percent_display_tasks(self.display_undispatched_count)

    @property
    def pct_display_tasks_failed(self) -> float:
        """
        Get the percentage of failed display tasks.

        :return: Percentage of failed display tasks.
        """
        return self._percent_display_tasks(self.display_failure_count)

    @property
    def pct_display_tasks_timed_out(self) -> float:
        """
        Get the percentage of timeout display tasks.

        :return: Percentage of timeout display tasks.
        """
        return self._percent_display_tasks(self.display_timed_out_count)

    @property
    def pct_display_tasks_system_failure(self) -> float:
        """
        Get the percentage of system failure display tasks.

        :return: Percentage of system failure display tasks.
        """
        return self._percent_display_tasks(self.display_system_failure_count)

    @property
    def create_time(self) -> Optional[datetime]:
        """
        Time the first task of the build was created.

        :return: Time first task was created.
        """
        if self._create_times:
            return min(self._create_times)
        return None

    @property
    def start_time(self) -> Optional[datetime]:
        """
        Time first task of build was started.

        :return: Time first task was started.
        """
        if self._start_times:
            return min(self._start_times)
        return None

    @property
    def end_time(self) -> Optional[datetime]:
        """
        Time last task of build was completed.

        :return: Time last task was completed.
        """
        if self._finish_times:
            return max(self._finish_times)
        return None

    @property
    def makespan(self) -> Optional[timedelta]:
        """
        Wall clock duration of build.

        :return: Timedelta duration of build.
        """
        if self.start_time and self.end_time:
            return self.end_time - self.start_time
        return None

    @property
    def wait_time(self) -> Optional[timedelta]:
        """
        Wall clock duration until build was started.

        :return: Timedelta duration until build was started.
        """
        if self.start_time and self.create_time:
            return self.start_time - self.create_time
        return None

    def _percent_tasks(self, n_tasks: int) -> float:
        """
        Calculate the percent of n_tasks out of total.

        :param n_tasks: Number of tasks to calculate percent of.
        :return: percentage n_tasks is out of total tasks.
        """
        if self.total_tasks == 0:
            return 0
        return n_tasks / self.total_tasks

    def _percent_display_tasks(self, n_tasks: int) -> float:
        """
        Calculate the percent of display n_tasks out of display total.

        :param n_tasks: Number of display tasks to calculate percent of.
        :return: percentage display n_tasks is out of total display tasks.
        """
        if self.total_display_tasks == 0:
            return 0
        return n_tasks / self.total_display_tasks

    def _count_task(self, task: "Task") -> None:
        """
        Add stats for the given task to the metrics.

        :param task: Task to add.
        """
        if task.is_undispatched():
            self.undispatched_count += 1
            if task.generated_by:
                self._display_map[task.generated_by].append(task)
            else:
                self.display_undispatched_count += 1

            return  # An 'undispatched' task has no useful stats.

        if task.is_active():
            LOGGER.warning("Active task found during metrics collection", task_id=task.task_id)
            raise ActiveTaskMetricsException(task, "Task in progress during metrics collection")

        if task.is_success():
            self.success_count += 1
            if not task.generated_by:
                self.display_success_count += 1
        else:
            self.failure_count += 1
            if not task.generated_by:
                self.display_failure_count += 1
            if task.is_system_failure():
                self.system_failure_count += 1
                if not task.generated_by:
                    self.display_system_failure_count += 1
            if task.is_timeout():
                self.timed_out_count += 1
                if not task.generated_by:
                    self.display_timed_out_count += 1

        if task.generated_by:
            self._display_map[task.generated_by].append(task)

        if task.ingest_time:
            self._create_times.append(task.ingest_time)
        else:
            self._create_times.append(task.start_time)

        if task.start_time:
            self._finish_times.append(task.finish_time)

        if task.start_time:
            self._start_times.append(task.start_time)

        self.total_processing_time += task.time_taken_ms / 1000

    def _count_display_tasks(self) -> None:
        for _, tasks in self._display_map.items():
            status = max([task.get_status_score() for task in tasks])
            if status == StatusScore.SUCCESS:
                self.display_success_count += 1
                continue
            if status == StatusScore.UNDISPATCHED:
                self.display_undispatched_count += 1
                continue
            if status == StatusScore.FAILURE:
                self.display_failure_count += 1
                continue
            if status == StatusScore.FAILURE_SYSTEM:
                self.display_system_failure_count += 1
                self.display_failure_count += 1
                continue
            if status == StatusScore.FAILURE_TIMEOUT:
                self.display_timed_out_count += 1
                self.display_failure_count += 1
                continue

    def as_dict(self, include_children: bool = False) -> Dict:
        """
        Provide a dictionary representation.

        :param include_children: Include child tasks in dictionary.
        :return: Dictionary of metrics.
        """
        metric = {
            "build": self.build.id,
            "total_processing_time": self.total_processing_time,
            "makespan": self.makespan.total_seconds() if self.makespan else None,
            "wait_time": self.wait_time.total_seconds() if self.wait_time else None,
            "total_tasks": self.total_tasks,
            "success_count": self.success_count,
            "pct_tasks_success": self.pct_tasks_success,
            "undispatched_count": self.undispatched_count,
            "pct_tasks_undispatched": self.pct_tasks_undispatched,
            "failure_count": self.failure_count,
            "pct_tasks_failed": self.pct_tasks_failed,
            "timed_out_count": self.timed_out_count,
            "system_failure_count": self.system_failure_count,
            "total_display_tasks": self.total_display_tasks,
            "success_display_count": self.display_success_count,
            "pct_display_tasks_success": self.pct_display_tasks_success,
            "undispatched_display_count": self.display_undispatched_count,
            "pct_display_tasks_undispatched": self.pct_display_tasks_undispatched,
            "failure_display_count": self.display_failure_count,
            "pct_display_tasks_failed": self.pct_display_tasks_failed,
            "timed_out_display_count": self.display_timed_out_count,
            "pct_display_tasks_timed_out": self.pct_display_tasks_timed_out,
            "system_failure_display_count": self.display_system_failure_count,
            "pct_display_tasks_system_failure": self.pct_display_tasks_system_failure,
        }

        if include_children:
            metric["tasks"] = [task.json for task in self.task_list]

        return metric

    def __str__(self) -> str:
        """
        Create string version of metrics.

        :return: String version of the metrics.
        """
        return """Build Id: {build_id}
        Total Processing Time: {total_processing_time:.2f}s ({total_processing_time_min:.2f}m)
        Makespan: {makespan:.2f}s ({makespan_min:.2f}m)
        Wait Time: {waittime:.2f}s ({waittime_min:.2f}m)
        Total Tasks: {total_tasks}
        Successful Tasks: {success_count} ({success_pct:.2%})
        Undispatched Tasks: {undispatched_count} ({undispatched_pct:.2%})
        Failed Tasks: {failed_count} ({failed_pct:.2%})
        Timeout Tasks: {timeout_count} ({timeout_pct:.2%})
        System Failure Tasks: {system_failure_count} ({system_failure_pct:.2%})
        Total Display Tasks: {total_display_tasks}
        Successful Display Tasks: {success_display_count} ({success_display_pct:.2%})
        Undispatched Display Tasks: {undispatched_display_count} ({undispatched_display_pct:.2%})
        Failed Display Tasks: {failed_display_count} ({failed_display_pct:.2%})
        Timeout Display Tasks: {timeout_display_count} ({timeout_display_pct:.2%})
        System Failure Tasks: {system_failure_display_count} ({system_failure_display_pct:.2%})
        """.format(
            build_id=self.build.id,
            total_processing_time=self.total_processing_time,
            total_processing_time_min=(self.total_processing_time / 60),
            makespan=(self.makespan.total_seconds()) if self.makespan else 0,
            makespan_min=(self.makespan.total_seconds() / 60) if self.makespan else 0,
            waittime=(self.wait_time.total_seconds()) if self.wait_time else 0,
            waittime_min=(self.wait_time.total_seconds() / 60) if self.wait_time else 0,
            total_tasks=self.total_tasks,
            success_count=self.success_count,
            success_pct=self.pct_tasks_success,
            undispatched_count=self.undispatched_count,
            undispatched_pct=self.pct_tasks_undispatched,
            failed_count=self.failure_count,
            failed_pct=self.pct_tasks_failed,
            timeout_count=self.timed_out_count,
            timeout_pct=self.pct_tasks_timed_out,
            system_failure_count=self.system_failure_count,
            system_failure_pct=self.pct_tasks_system_failure,
            total_display_tasks=self.total_display_tasks,
            success_display_count=self.display_success_count,
            success_display_pct=self.pct_display_tasks_success,
            undispatched_display_count=self.display_undispatched_count,
            undispatched_display_pct=self.pct_display_tasks_undispatched,
            failed_display_count=self.display_failure_count,
            failed_display_pct=self.pct_display_tasks_failed,
            timeout_display_count=self.display_timed_out_count,
            timeout_display_pct=self.pct_display_tasks_timed_out,
            system_failure_display_count=self.display_system_failure_count,
            system_failure_display_pct=self.pct_display_tasks_system_failure,
        ).rstrip()



# -*- encoding: utf-8 -*-
"""Version representation of evergreen."""
from __future__ import absolute_import

from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional

import structlog

from evergreen.base import _BaseEvergreenObject, evg_attrib, evg_datetime_attrib
from evergreen.build import Build
from evergreen.manifest import ManifestModule
from evergreen.metrics.versionmetrics import VersionMetrics

if TYPE_CHECKING:
    from evergreen.api import EvergreenApi
    from evergreen.manifest import Manifest
    from evergreen.patch import Patch  # noqa: F401


LOGGER = structlog.getLogger(__name__)


class Requester(str, Enum):
    """Requester that created version."""

    PATCH_REQUEST = "patch_request"
    GITTER_REQUEST = "gitter_request"
    GITHUB_PULL_REQUEST = "github_pull_request"
    MERGE_TEST = "merge_test"
    AD_HOC = "ad_hoc"
    TRIGGER_REQUEST = "trigger_request"
    UNKNOWN = "UNKNOWN"

    def evg_value(self) -> str:
        """Get the evergreen value for a requester."""
        return self.name.lower()

    def stats_value(self) -> str:
        """Get the value for the stats endpoints."""
        value_mappings = {
            Requester.PATCH_REQUEST: "patch",
            Requester.GITTER_REQUEST: "mainline",
            Requester.GITHUB_PULL_REQUEST: "patch",
            Requester.MERGE_TEST: "",
            Requester.AD_HOC: "adhoc",
            Requester.TRIGGER_REQUEST: "trigger",
            Requester.UNKNOWN: "",
        }

        return value_mappings[self]


PATCH_REQUESTERS = {
    Requester.PATCH_REQUEST,
    Requester.GITHUB_PULL_REQUEST,
    Requester.MERGE_TEST,
}

EVG_VERSION_STATUS_SUCCESS = "success"
EVG_VERSION_STATUS_FAILED = "failed"
EVG_VERSION_STATUS_CREATED = "created"

COMPLETED_STATES = {
    EVG_VERSION_STATUS_FAILED,
    EVG_VERSION_STATUS_SUCCESS,
}


class BuildVariantStatus(_BaseEvergreenObject):
    """Representation of a Build Variants status."""

    build_variant = evg_attrib("build_variant")
    build_id = evg_attrib("build_id")

    def __init__(self, json: Dict[str, Any], api: "EvergreenApi") -> None:
        """Create an instance of a Build Variants status."""
        super(BuildVariantStatus, self).__init__(json, api)

    def get_build(self) -> "Build":
        """Get the build object for this build variants status."""
        return self._api.build_by_id(self.build_id)


class Version(_BaseEvergreenObject):
    """Representation of an Evergreen Version."""

    version_id = evg_attrib("version_id")
    create_time = evg_datetime_attrib("create_time")
    start_time = evg_datetime_attrib("start_time")
    finish_time = evg_datetime_attrib("finish_time")
    revision = evg_attrib("revision")
    order = evg_attrib("order")
    project = evg_attrib("project")
    author = evg_attrib("author")
    author_email = evg_attrib("author_email")
    message = evg_attrib("message")
    status = evg_attrib("status")
    repo = evg_attrib("repo")
    branch = evg_attrib("branch")
    errors = evg_attrib("errors")
    warnings = evg_attrib("warnings")
    ignored = evg_attrib("ignored")
    project_identifier = evg_attrib("project_identifier")
    aborted = evg_attrib("aborted")
    parameters = evg_attrib("parameters")

    def __init__(self, json: Dict[str, Any], api: "EvergreenApi") -> None:
        """
        Create an instance of an evergreen version.

        :param json: json representing version
        """
        super(Version, self).__init__(json, api)

        if "build_variants_status" in self.json and self.json["build_variants_status"]:
            self.build_variants_map = {
                bvs["build_variant"]: bvs["build_id"] for bvs in self.json["build_variants_status"]
            }
            LOGGER.debug(
                "build_variants_map initialized for version",
                version_id=self.version_id,
                build_variants_map=self.build_variants_map,
            )
        else:
            LOGGER.debug(
                "build_variants_status either empty or not found for version",
                version_id=self.version_id,
                json=self.json,
            )
            self.build_variants_map = {}

    @property
    def build_variants_status(self) -> List[BuildVariantStatus]:
        """Get a list of build variant statuses."""
        if "build_variants_status" not in self.json or not self.json["build_variants_status"]:
            return []
        build_variants_status = self.json["build_variants_status"]
        return [BuildVariantStatus(bvs, self._api) for bvs in build_variants_status]

    @property
    def requester(self) -> Requester:
        """Get the requester of this version."""
        return Requester[self.json.get("requester", "UNKNOWN").upper()]

    def build_by_variant(self, build_variant: str) -> "Build":
        """
        Get a build object for the specified variant.

        :param build_variant: Build variant to get build for.
        :return: Build object for variant.
        """
        return self._api.build_by_id(self.build_variants_map[build_variant])

    def get_manifest(self) -> "Manifest":
        """
        Get the manifest for this version.

        :return: Manifest for this version.
        """
        return self._api.manifest(self.project, self.revision)

    def get_modules(self) -> Optional[Dict[str, ManifestModule]]:
        """
        Get the modules for this version.

        :return: ManifestModules for this version.
        """
        return self.get_manifest().modules

    def get_builds(self) -> List["Build"]:
        """
        Get all the builds that are a part of this version.

        :return: List of build that are a part of this version.
        """
        return self._api.builds_by_version(self.version_id)

    def is_patch(self) -> bool:
        """
        Determine if this version from a patch build.

        :return: True if this version is a patch build.
        """
        if self.requester and self.requester != Requester.UNKNOWN:
            return self.requester in PATCH_REQUESTERS
        return not self.version_id.startswith(self.project.replace("-", "_"))

    def is_completed(self) -> bool:
        """
        Determine if this version has completed running tasks.

        :return: True if version has completed.
        """
        return self.status in COMPLETED_STATES

    def get_patch(self) -> Optional["Patch"]:
        """
        Get the patch information for this version.

        :return: Patch for this version.
        """
        if self.is_patch():
            return self._api.patch_by_id(self.version_id)
        return None

    def get_metrics(self, task_filter_fn: Optional[Callable] = None) -> Optional[VersionMetrics]:
        """
        Calculate the metrics for this version.

        Metrics are only available on versions that have finished running.

        :param task_filter_fn: function to filter tasks included for metrics, should accept a task
                               argument.
        :return: Metrics for this version.
        """
        if self.status != EVG_VERSION_STATUS_CREATED:
            return VersionMetrics(self).calculate(task_filter_fn)
        return None

    def __repr__(self) -> str:
        """
        Get the string representation of Version for debugging purposes.

        :return: String representation of Version.
        """
        return "Version({id})".format(id=self.version_id)


class RecentVersionRow(_BaseEvergreenObject):
    """Wrapper for a row of the RecentVersions endpoint."""

    build_variant = evg_attrib("build_variant")

    @property
    def builds(self) -> Dict[str, Build]:
        """Get a map of build IDs to build objects."""
        return {k: Build(v, self._api) for k, v in self.json["builds"].items()}


class RecentVersions(_BaseEvergreenObject):
    """Wrapper for the data object returned by /projects/{project_id}/recent_versions."""

    rows = evg_attrib("rows")
    build_variants = evg_attrib("build_variants")

    @property
    def row_map(self) -> Dict[str, RecentVersionRow]:
        """Get a map of build names to RecentVersionRows."""
        return {k: RecentVersionRow(v, self._api) for k, v in self.json["rows"].items()}

    @property
    def versions(self) -> List[Version]:
        """
        Get the list of versions from the recent versions response object.

        :return: List of versions from the response object
        """
        return [Version(wrapper["versions"], self._api) for wrapper in self.json["versions"]]



"""Models to working with task annotations."""
from typing import TYPE_CHECKING, Any, Dict, List

from evergreen.base import _BaseEvergreenObject, evg_attrib, evg_datetime_attrib

if TYPE_CHECKING:
    from evergreen.api import EvergreenApi


class Source(_BaseEvergreenObject):
    """Source of where an annotation was generated."""

    author = evg_attrib("author")
    time = evg_datetime_attrib("time")
    requester = evg_attrib("requester")

    def __init__(self, json: Dict[str, Any], api: "EvergreenApi") -> None:
        """Create an instance of a task annotation source."""
        super().__init__(json, api)


class IssueLink(_BaseEvergreenObject):
    """Representation of a issue added as a task annotation."""

    url = evg_attrib("url")
    issue_key = evg_attrib("issue_key")

    def __init__(self, json: Dict[str, Any], api: "EvergreenApi") -> None:
        """Create an instance of a task annotation issue link."""
        super().__init__(json, api)

    @property
    def source(self) -> Source:
        """Get the source of this issue link."""
        return Source(self.json["source"], self._api)


class Note(_BaseEvergreenObject):
    """Representation of a note associated with a task annotation."""

    message = evg_attrib("message")

    def __init__(self, json: Dict[str, Any], api: "EvergreenApi") -> None:
        """Create an instance of a task annotation note."""
        super().__init__(json, api)

    @property
    def source(self) -> Source:
        """Get the source of this note."""
        return Source(self.json["source"], self._api)


class MetadataLink(_BaseEvergreenObject):
    """Representation of a metadata link associated with a task annotation."""

    url = evg_attrib("url")
    text = evg_attrib("text")

    def __init__(self, json: Dict[str, Any], api: "EvergreenApi") -> None:
        """Create an instance of a task annotation metadata link."""
        super().__init__(json, api)

    @property
    def source(self) -> Source:
        """Get the source of this metadata link."""
        return Source(self.json["source"], self._api)


class TaskAnnotation(_BaseEvergreenObject):
    """Representation of a task annotation."""

    task_id = evg_attrib("task_id")
    task_execution = evg_attrib("task_execution")

    def __init__(self, json: Dict[str, Any], api: "EvergreenApi") -> None:
        """Create an instance of a task annotations."""
        super().__init__(json, api)

    @property
    def issues(self) -> List[IssueLink]:
        """Get the issues this task has been annotated with."""
        return [IssueLink(j, self._api) for j in self.json.get("issues", [])]

    @property
    def suspected_issues(self) -> List[IssueLink]:
        """Get the suspected issues this task has been annotated with."""
        return [IssueLink(j, self._api) for j in self.json.get("suspected_issues", [])]

    @property
    def metadata(self) -> Dict[str, Any]:
        """Get metadata associated with this annotation."""
        return self.json.get("metadata", {})

    @property
    def note(self) -> Note:
        """Get a note about this annotation."""
        return Note(self.json.get("note", {}), self._api)

    @property
    def metadata_links(self) -> List[MetadataLink]:
        """Get metadata links for this annotation."""
        return [MetadataLink(link, self._api) for link in self.json.get("metadata_links", [])]
